/* 
 * tplay - buffered audio player  
 *
 * (c) 1997 ilkka karvinen <ik@iki.fi>
 *
 * Copyright under the GNU GENERAL PUBLIC LICENSE
 *   (see the file COPYING in this directory)
 *
 */

#include "common.h"
#include "audio.h"
#include <getopt.h>
#include <pthread.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

/* Local functions */
static void init_info(char **argv);
static void reset_info();
static void allocate_buffer();
static void calculate_buffer_size(int speed, int bits, int channels);
static void calculate_skip_size(int speed, int bits, int channels);
static int  buffer_usage();
static void print_usage(int state);
static void read_header();
static void fill_buffer();
static void swap_block(byte *buffer, int blocksize);
static void producer();
static void *consumer(void *data);
static void play(char *file);
static void decode_env_switches();
static void decode_switches (int argc, char **argv);

#define OPTION_STRING "D:B:b:hms:S:e:uvVxrfl"

static struct option long_options[] =
{ { "buffer-size", required_argument, NULL, 'B' },
  { "bits", required_argument, NULL, 'b' },
  { "encoding", required_argument, NULL, 'e' },
  { "help", no_argument, NULL, 'h' },
  { "mono", no_argument, NULL, 'm' },
  { "raw", no_argument, NULL, 'r' },
  { "skip", required_argument, NULL, 'S' },
  { "speed", required_argument, NULL, 's' },
  { "usage", no_argument, NULL, 'u' },
  { "version", no_argument, NULL, 'v' },
  { "swap", no_argument, NULL, 'x' },
  { "verbose", no_argument, NULL, 'V' },
  { "force", no_argument, NULL, 'f' },
  { "device", required_argument, NULL, 'D' },
  { "loop", no_argument, NULL, 'l' },
  { 0, 0, 0, 0 } };

struct info_struct info;

static pthread_cond_t  cond;        /* Condition variable */
static pthread_mutex_t lock;        /* Mutex lock */

/* Initialize the buffer information */
static void init_info(char **argv)
{
  info.progname = PACKAGE;
  info.loop = FALSE;
  info.skip = FALSE;
  info.in_seconds = FALSE;
  info.speed    = DEFAULT_SPEED;
  info.bits     = DEFAULT_BITS;
  info.encoding = LIN_BITS;
  info.channels = DEFAULT_CHANNELS;
  info.buffer_size = BUFFER_SIZE;
  info.show_usage = FALSE;
  info.swap = FALSE;
  info.forceraw = FALSE;
  info.force = FALSE;
  info.device = NULL;
  info.verbose = FALSE;
  info.optind = 0;
  info.buffer = NULL;
  info.firstblock = NULL;
  info.number_of_blocks = 0;
  info.alldone = FALSE;
  info.audioset = FALSE;
  info.headerskip = 0;
  info.audio_fd = -1;
  info.file_fd = -1;
  info.blocksize = 0;
  info.bytes_on_last_block = 0;
  info.capacity = 0;
  info.writeblock = 0;
  info.data_length = 0;
  pthread_mutex_init(&lock, NULL);
  pthread_cond_init(&cond, NULL);
}

static void reset_info()
{
  info.firstblock = NULL;
  info.writeblock = 0;
  info.capacity = 0;
  info.alldone = FALSE;
  info.audioset = FALSE;
  info.headerskip = 0;
  info.bytes_on_last_block = 0;
  pthread_mutex_init(&lock, NULL);
  pthread_cond_init(&cond, NULL);
}

static void allocate_buffer()
{
  if (info.buffer == NULL) {
    info.buffer = (byte*) malloc(info.buffer_size*sizeof(byte));
    if (info.buffer == NULL) 
      errdie("malloc failed");
    info.number_of_blocks = info.buffer_size / info.blocksize;
  }
}

/* Buffer usage in percentage */
static int buffer_usage()
{
  int usage;
  pthread_mutex_lock(&lock);
  usage = (100 * (info.capacity)) / 
              info.number_of_blocks;
  pthread_mutex_unlock(&lock);
  return(usage);
}

static void calculate_buffer_size(int speed, int bits, int channels)
{
  int bytes_per_second;
  double seconds = 0.0;
  int tmp;

  bytes_per_second = speed * channels * (bits / 8);

  if (info.in_seconds) {
    tmp = (int)(1000.0 * info.seconds);
    info.buffer_size = bytes_per_second * tmp / 1000;
    seconds = info.seconds;
  } else {
    seconds = (double)info.buffer_size / (double)bytes_per_second;
  }

  if (info.buffer_size < 5*MIN_BLOCK_SIZE)
    info.buffer_size = 5*MIN_BLOCK_SIZE;

  if (info.verbose) 
    printf("Buffer size: %dk (%.1fs)\n", info.buffer_size / 1024, seconds);

}

static void calculate_skip_size(int speed, int bits, int channels)
{
  int bytes_per_second;

  if (info.skip) {
    bytes_per_second = speed * channels * (bits / 8);
    info.skip_size = bytes_per_second * info.sseconds;
    if (info.verbose) 
      printf("Skipping: %dk (%.1fs)\n", info.skip_size / 1024, info.sseconds);
    if (lseek(info.file_fd, (off_t)info.skip_size, SEEK_CUR) == (off_t)-1)
      warning("Failed to skip. Behavior may be erratic. Continuing anyway.");
  }
}

static void print_usage(int state)
{
  static int prev_usage = -1;
  int usage;

  if (state != 0) {
    if ((usage = buffer_usage()) != prev_usage) {
      printf("Buffer: %d%%   \r", buffer_usage());
      fflush(stdout);
      prev_usage = usage;
    }
  } else {
    printf("Buffer: rebuf    \r");
    fflush(stdout);
  }
}

static void read_header()
{
  int bytesread, count;
  byte *p, *bufferp;

  info.firstblock = (byte*) malloc(info.blocksize*sizeof(byte));
  if (info.firstblock == NULL) 
    errdie("malloc failed");

  /* Set bufferp to point to the beginning of buffer */
  bufferp = info.firstblock;
  /* If raw format is not forced, check header */
  if (!info.forceraw) {
    /* Read one block for header */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, 
			  info.blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }

    /* If less than about a Sun audio header read,     */
    /* the sample is too small.                        */
    if (bytesread < SUN_HDRSIZE)
      die("Sample size is too small");

    /* Check Sun and RIFF/WAVE audio headers */
    if (read_au(info.firstblock) && read_wav(info.firstblock)) {
      if (info.verbose)
	printf("Playing raw audio data: \
%d samples/s, %d bits, %d channel(s).\n", 
	       info.speed, info.bits, info.channels);
    } 

    /* Swap the block if needed. The header is swapped also but    */
    /* it is not needed anymore.                                   */
    if (info.swap)
      swap_block(bufferp, bytesread);

    /* If less bytes read than block size => all read => quit      */
    if (bytesread < info.blocksize) {
      info.alldone = TRUE;
      info.bytes_on_last_block = bytesread;
    }

    /* If header was found, read to the end of the block.          */
    if (info.headerskip) {
      count = 0;
      bytesread = info.blocksize - info.headerskip; 
      p = info.firstblock + (info.blocksize - info.headerskip);
      while ((bytesread < info.blocksize) && (count != -1) &&
	     ((count = read(info.file_fd, p, 
			    info.blocksize - bytesread)) != 0))
	{
	  p+=count; bytesread+=count;
	}
    }

    /* Advance to the next block */
    info.writeblock++;
    info.capacity++;

    /* We know the length of the data. Use it. */
    if (info.data_length) {
      if (info.data_length > info.blocksize)
	info.data_length -= info.blocksize;
      else
	{
	  /* We've already read all the data */
	  info.alldone = TRUE;
	  info.bytes_on_last_block = info.data_length;
	}
    }

  } else /* Raw audio mode forced */
    if (info.verbose)
      printf("Playing raw audio data: \
%d samples/s, %d bits, %d channel(s).\n", 
	     info.speed, info.bits, info.channels);

  calculate_buffer_size(info.speed, info.bits, info.channels);
  allocate_buffer();
  calculate_skip_size(info.speed, info.bits, info.channels);
}

static void fill_buffer()
{
  int bytesread, count;
  byte *p, *bufferp;

  /* Move headerblock to buffer first */
  memmove(info.buffer, info.firstblock, info.blocksize);
  /* First block is not needed anymore so we free it */
  free(info.firstblock);

  bufferp = info.buffer + info.blocksize;
  while(TRUE) {

    /* Read a block piece by piece; especially if read from     */
    /* standard input, read may return unwanted block size.     */
    /* If count is -1, file_fd is invalid.                      */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < info.blocksize) && (count != -1) &&
	   ((count = read(info.file_fd, p, 
			  info.blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }

    if (info.swap)
      swap_block(bufferp, bytesread);

    if (info.data_length) {
	if (info.data_length > info.blocksize)
	    info.data_length -= info.blocksize;
	else
	{
	    /* We've already read all the data */
	    info.last_block = info.writeblock;
	    info.bytes_on_last_block = info.data_length;
	    info.alldone = TRUE;
	    break;
	}
    }

    if (bytesread < info.blocksize) {
      info.last_block = info.writeblock;
      info.bytes_on_last_block = bytesread;
      info.alldone = TRUE;
      break;
    }

    /* Advance some pointers */
    bufferp+=info.blocksize;
    info.writeblock++;
    info.capacity++;
    if (info.writeblock >= info.number_of_blocks) {
      info.writeblock = 0;
      break;
    }
  }
}

static void swap_block(byte *buffer, int blocksize)
{
  register int i;
  byte c, *p;
  p = buffer;
  for (i=0; i<blocksize/2; i++) {
    c = *p;
    *p = *(p+1);
    *++p = c;
    p++;
  }
}

static void producer()
{
  int bytesread, count, writeblock;
  int overflow, file_fd, blocksize;
  byte *p, *bufferp, *bufferbase;
  int number_of_blocks, show_usage;

  /* Endless loop for the audio file reader and bufferer. */
  pthread_mutex_lock(&lock);
  bufferp = bufferbase = info.buffer;
  file_fd = info.file_fd;
  blocksize = info.blocksize;
  number_of_blocks = info.number_of_blocks;
  writeblock = info.writeblock;
  show_usage = info.show_usage;
  pthread_mutex_unlock(&lock);
  for ( ; ; ) {
    pthread_mutex_lock(&lock);
    overflow = (info.capacity >= info.number_of_blocks) ? TRUE : FALSE;
    pthread_mutex_unlock(&lock);

    if (overflow) {
      /* Overflow: Buffer is full */
      /* usleep(10000); Take a short nap or we will piss off CPU */
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 10000;
      select(0, NULL, NULL, NULL, &tv);
      continue; /* Loop back and try again */
    }

    /* Read a block piece by piece; especially if read from     */
    /* standard input, read may return unwanted block size.     */
    /* If count is -1, file_fd is invalid.                      */
    bytesread = 0; count = 0;
    p = bufferp;
    while ((bytesread < blocksize) && (count != -1) &&
	   ((count = read(file_fd, p, 
			  blocksize - bytesread)) != 0))
      {
	p+=count; bytesread+=count;
      }
    /*    printf("in: %d\n", writeblock);*/

    if (info.swap)
      swap_block(bufferp, bytesread);

    /* Use the length of the data if set. */
    if (info.data_length) {
      if (info.data_length > info.blocksize)
	info.data_length -= info.blocksize;
      else
	{
	  /* We've already read all the data */
	  pthread_mutex_lock(&lock);
	  info.last_block = writeblock;
	  info.bytes_on_last_block = info.data_length;
	  info.alldone = TRUE;
	  pthread_mutex_unlock(&lock);
	  break;
	}
    }

    /* If less than 'blocksize' read, we assume that this is          */
    /* the last block. Save data for consumer and exit this function. */
    /* If bytesread is -1, file_fd was invalid.                       */
    if (bytesread < blocksize) {
      pthread_mutex_lock(&lock);
      info.last_block = writeblock;
      info.bytes_on_last_block = bytesread;
      info.alldone = TRUE;
      pthread_mutex_unlock(&lock);
      break;
    }

    /* Advance some pointers */
    bufferp+=blocksize;
    writeblock++;
    if (writeblock >= number_of_blocks) {
      bufferp = bufferbase;
      writeblock = 0;
    }

    /* Increase buffer capacity */
    pthread_mutex_lock(&lock);
    info.capacity++;
    pthread_mutex_unlock(&lock);

  }

}

static void *consumer(void *data)
{
  byte *bufferp;
  byte *bufferbase;
  int retry, readblock, number_of_blocks;
  int audio_fd, blocksize, underflow;
  int lastblock, alldone, show_usage;

  /* Endless loop for the audio player */
  pthread_mutex_lock(&lock);
  audio_fd = info.audio_fd;
  blocksize = info.blocksize;
  number_of_blocks = info.number_of_blocks;
  bufferp = bufferbase = info.buffer;
  show_usage = info.show_usage;
  pthread_mutex_unlock(&lock);
  readblock = 0;
  for ( ; ; ) {
    if (show_usage)
      print_usage(1);

    pthread_mutex_lock(&lock);
    alldone = info.alldone;
    underflow = (info.capacity <= 0) ? TRUE : FALSE;
    lastblock = (info.last_block == readblock) ? TRUE : FALSE;
    audio_fd = info.audio_fd;
    pthread_mutex_unlock(&lock); 
   
    if (alldone) {
      /* If we hit the last block, write the last bytes and exit   */
      if (lastblock) {
	/* Write the last bytes and exit */
	/* info is not guarded by mutex here because producer      */
	/* should be exit by now.                                  */
	write(audio_fd, bufferp, info.bytes_on_last_block);
	break;
      }
    }

    if (underflow) {
      /* The audio buffer is empty; give producer a break */
      if (show_usage) 
	print_usage(0);
      sync_audio();
      sleep(1);
      continue;
    }

    /* Push the audio data block to device */
    write(audio_fd, bufferp, blocksize);
    /*    printf("out: %d\n", readblock);*/

    /* Advance some pointers */
    bufferp+=blocksize;
    readblock++;
    if (readblock >= number_of_blocks) {
      bufferp = bufferbase;
      readblock = 0;
    }

    /* Decrease buffer capacity */
    pthread_mutex_lock(&lock);
    info.capacity--;
    pthread_mutex_unlock(&lock);

  }

  return NULL;
}

void play(char *file)
{
  pthread_attr_t attr;
  pthread_t thread;
  void *retval;
  int err;

  if (info.verbose) {
    if (file != NULL)
      printf("Playing: %s\n", file);
    else
      printf("Playing from standard input.\n");
  }

  /* Reset flags */
  reset_info();

  open_audio();

  /* If filename is given, open it. Otherwise, stdin is used. */
  if (file != NULL)
    open_file(file);

  read_header();

  /* Set audio parameters */
  set_audio_parameters();

  /* Fill the audio buffer */
  fill_buffer();

  /* Set thread attributes: FIFO. */
  pthread_attr_init(&attr);
  pthread_attr_setschedpolicy(&attr, SCHED_FIFO);

  if ( (err=pthread_create(&thread, &attr, consumer, 0)) ) {
    /* From Edward A. Falk: Only superuser can set SCHED_FIFO */
    /* so try pthread_create without attributes if failed.    */
    if( err != EPERM || pthread_create(&thread, NULL, consumer, 0) )
      errdie("Cannot create thread");
    warning("No permissions for priority scheduling\n");
  }

  /* Start producing. Don't start if sample was smaller than  */
  /* audio buffer size (alldone flag is set)                  */
  if (!info.alldone)
    producer();

  /* Wait until the thread has finished. */
  pthread_join(thread, &retval);

  /* Close input file */
  close_file();
  close_audio();
}

static void decode_env_switches()
{
  char *p;

  /* If audio device is given in environment variable */
  /* TPLAYDEV, use it.                                */
  if ((p = getenv("TPLAYDEV"))) {
    info.device = (char*) malloc(strlen(p)+1);
    if (info.device != NULL)
      strcpy(info.device, p);
  }
#if defined(sun)
  /* A property of "well-behaved" Sun audio apps */
  else {
    if ((p = getenv("AUDIODEV"))) {
      info.device = p;
    }
  }
#endif
}

static void decode_switches(int argc, char **argv)
{
  int optc;

  while (optc = getopt_long (argc, argv, OPTION_STRING, long_options, NULL),
	 optc != EOF)
    {
      switch(optc)
	{
	case 's' :
	  info.speed = atoi(optarg);
	  break;
	case 'e' :
	  /* bits of course needs to match... */
	  if (!strcasecmp(optarg, "linear")) {
	    info.encoding = LIN_BITS;
	    break;
	  }
	  if (!strcasecmp(optarg, "ulaw")) {
	    info.encoding = ULAW_BITS;
	    break;
	  }
	  if (!strcasecmp(optarg, "linear8")) {
	    info.encoding = LIN8_BITS;
	    break;
	  }
	  if (!strcasecmp(optarg, "alaw")) {
	    info.encoding = ALAW_BITS;
	    break;
	  }
	  if (!strcasecmp(optarg, "adpcm")) {
	    info.encoding = ADPCM_BITS;
	    break;
	  }
	  fprintf(stderr, "Unknown encoding type %s. Continuing anyway.\n", optarg);
	  break;
	case 'b' :
	  info.bits = atoi(optarg);
	  break;
	case 'm' :
	  info.channels = 1;
	  break;
	case 'S' :
	  info.skip = TRUE;
	  info.sseconds = atof(optarg);
	  break;
	case 'B' :
	  if (strchr(optarg, '.')) {
	    info.in_seconds = TRUE;
	    info.seconds = atof(optarg);
	  } else {
	    info.in_seconds = FALSE;
	    info.buffer_size = 1024*atoi(optarg);
	  }
	  break;
	case 'u' :
	  info.show_usage = TRUE;
	  break;
	case 'v' :
	  printf("%s version %s\n", PACKAGE, VERSION); 
	  exit(-1);
	  break;
	case 'V' :
	  info.verbose = TRUE;
	  break;
	case 'x' :
	  info.swap = TRUE;
	  break;
	case 'r' :
	  info.forceraw = TRUE;
	  break;
	case 'f' :
	  info.force = TRUE;
	  break;
	case 'D' :
	  info.device = (char*) malloc(strlen(optarg)+1);
	  if (info.device != NULL)
	    strcpy(info.device, optarg);
	  break;
	case 'l' :
	  info.loop = TRUE;
	  break;
	case 'h' :
	  usage();
	  break;
	default:
	  usage();
	}
    }

  /* Check if filename is given. If not, use standard input. */
  if (optind > argc - 1) 
    info.file_fd = STDIN_FILENO; 

  info.optind = optind;
}

int main(int argc, char *argv[])
{
  int i;

  /* Init info structure */
  init_info(argv);
  
  /* Decode first environment variables and then command line */
  /* options.                                                 */
  decode_env_switches();
  decode_switches(argc, argv);

  /* Play file */
  if (info.file_fd == STDIN_FILENO) {
    play(NULL);
  } else {
    do {
      for (i=info.optind; i<argc; i++)
	play(argv[i]);
    } while(info.loop);
  }
    
  destroy_buffer();

  return 0;
}
  


